

// Schedules for irregular subscripting of distributed arrays.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "platform.h"
#include "globals.h"
#include "ProcTree.h"
#include "Procs.h"
#include "Range.h"
#include "Frame.h"

#include "raw_type.h"

#include "interface.h"

extern "C" {
  void nplc_tsc(MPI_Comm  WORLD_input,
                void *source_array,
		int source_lidx[],
		int NP,                       /* s_L_all */
		int s_N_all_input,            /* s_N_all */

		int index_lidx[],
		int index_pnum[],

		void *target_array,
		int MP,                      /* t_L_all */
					     /* All proc. have the same MP */
		Data_Types data_type_input,
		Operator_Types operator_type_input) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `NPLCSkeleton'                                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

NPLCSkeleton :: NPLCSkeleton(const Frame* _loc, const Frame* _rem,
                             const Frame** _subs, int** subsData,
                             const Frame* _msk, int* mskData,
			     Data_Types _data_type_input,
			     Operator_Types _operator_type_input) :
      loc(_loc), rem(_rem), 
      subs(_subs), msk(_msk), rank(_loc->rnk()),
      data_type_input(_data_type_input),
      operator_type_input(_operator_type_input) {

  int r ;

  // Process id translation tables.

  // Translation between id relative to active process group and id relative
  // to initial process group (absolute id).

  int nini ;
  MPI_Comm_size(MPI_COMM_WORLD, &nini) ;
  int napg = apg.size() ;

  apg2ini = new int [napg] ;
  ini2apg = new int [nini] ;

  for(int i = 0 ; i < napg ; i++) {
    int ini = apg.id_abs(i) ;

    apg2ini [i]   = ini ;
    ini2apg [ini] = i ;
  }
  
  // MPI communicator describing active process group.

  MPI_Group ini_group ;
  MPI_Comm_group(MPI_COMM_WORLD, &ini_group) ;

  MPI_Group_incl(ini_group, napg, apg2ini, &apg_group) ;
  MPI_Comm_create(MPI_COMM_WORLD, apg_group, &apg_comm) ;

  MPI_Group_free(&ini_group) ;

  // Strides of immediate target array and bound on number of local elements.

  tar_strs = new Stride [rem->rnk()] ;
  MP = 1 ;
  for(r = 0 ; r < rem->rnk() ; r++)
    tar_strs [r] = rem->rng(r).stride(MP) ;

  // Bound on number of all elements of source array and local elements
  // of source array

  s_N_all_input = 1 ;
  int NP_bound  = 1 ;
  for(r = 0 ; r < loc->rnk() ; r++) {
    s_N_all_input *= loc->rng(r).size() ;     // ignores masking
    NP_bound      *= loc->rng(r).volume() ;
  }

  // Vectors of source and destination offsets

  source_lidx = new int [NP_bound] ;
                                             
  index_lidx  = new int [NP_bound] ;
  index_pnum  = new int [NP_bound] ;
  
  NP = 0 ;

  Group p = loc->grp() ;
  if(p.on()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.
    rep.where() ;
      if(apg.lead_prc() == p.lead_prc())
	loop(0, subsData, mskData, 0) ;
    rep.erewh() ;
  } p.no() ;
}

NPLCSkeleton :: ~NPLCSkeleton() {
  delete [] index_pnum ;
  delete [] index_lidx ;

  delete [] source_lidx ;

  delete [] tar_strs ;

  MPI_Comm_free(&apg_comm) ;
  MPI_Group_free(&apg_group) ;

  delete [] ini2apg ;
  delete [] apg2ini ;
}

void NPLCSkeleton :: loop(int locOffset, int** subsElem, int* mskElem,
                          const int r) {

  // Multi-dimensional loop to traverse elements of the subscript arrays

  if(r == rank) {
    if(!msk || *mskElem) {
      Group  remGrp = rem->grp() ;
      int remOffset = 0 ;

      for(int s = 0 ; s < rem->rnk() ; s++) {
	Range y(rem->rng(s)) ;
	Subscript i(y, *(subsElem [s])) ;

        remOffset += tar_strs [s].offset(i) ;
        remGrp    /= i ;
      }

      source_lidx [NP] = locOffset ;
                                             
      index_lidx  [NP] = remOffset ;
      index_pnum  [NP] = ini2apg [remGrp.lead_abs()] ;
             // (Must convert absolute id to id relative to `apg')

      NP++ ;
    }
  }
  else {
    LocBlocksIndex i(loc->rng(r)) ;

    Stride locStr = loc->str(r) ;
    Stride mskStr ;
    if(msk) mskStr = msk->str(r) ;

    int** subsDataElem = new int* [rem->rnk()] ;
    int* subsStep = new int [rem->rnk()] ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      int s ;

      int locOffsetElem = locOffset + locStr.offset(i) ;
      const int locStep = locStr.step(i) ;

      for(s = 0 ; s < rem->rnk() ; s++) {
        Stride subStr = (subs [s])->str(r) ;

        subsDataElem [s] = subsElem [s] + subStr.offset(i) ;
        subsStep [s]     = subStr.step(i) ;
      }

      int* mskDataElem  = msk ? mskElem + mskStr.offset(i) : (int*) NULL ;
      const int mskStep = msk ? mskStr.step(i) : 0 ;

      for(int i_block = 0 ; i_block < i.count ; i_block++) {
        loop(locOffsetElem, subsDataElem, mskDataElem, r + 1) ;

	locOffsetElem += locStep ;
        for(s = 0 ; s < rem->rnk() ; s++)
          subsDataElem [s] += subsStep [s] ;
        if(msk)
          mskDataElem += mskStep ;
      }
    }

    delete [] subsStep ;
    delete [] subsDataElem ;
  }
}

void NPLCSkeleton :: execute(void* source_array, void* target_array) {
  nplc_tsc(apg_comm,
           source_array, source_lidx, NP, s_N_all_input,
           index_lidx, index_pnum, target_array, MP,
           data_type_input, operator_type_input) ;
}

